/*==============================================================================
   DEMO.C :- FORM Demo Program (MS Windows Version)
   Sub Systems, Inc.
   Software License Agreement  (1989)
   ----------------------------------
   This is a demo program.  The program shows how you can utilize the Report
   Ease routines to perform develop report templates and execute the reports.
===============================================================================*/

#include "windows.h"
#include "stdio.h"

#ifdef SSMAPLEAK
  #define _CRTDBG_MAP_ALLOC
  #include <stdlib.h>
  #include <crtdbg.h>
#else
  #include "stdlib.h"
#endif

#include "fcntl.h"
#include "sys\types.h"
#include "sys\stat.h"
#include "io.h"
#include "dos.h"
#include "string.h"
#include "math.h"

#if defined(__TURBOC__)
   #include "dir.h"
   #include "mem.h"
#else
   #include "memory.h"
#endif

//#define CHARTPRO_INTERFACE    // show an example of interfacing with ChartPro
#if defined(CHARTPRO_INTERFACE)
   #include "c:\chc\chart.h"
#endif

/******************************************************************************
    WIN32 specific defines
*******************************************************************************/
#if defined (_WIN32)
   #if !defined(WIN32)
     #define WIN32
   #endif
#endif
#if defined (WIN32)
   #undef  huge
   #define huge
   #undef  FreeProcInstance
   #define FreeProcInstance(x)
   #define COMMAND_ID(wp,lp)        LOWORD(wp)
   #define CONTROL_ID(wp,lp)        LOWORD(wp)
   #define NOTIFY_MSG(wp,lp)        HIWORD(wp)
   #undef  farmalloc
   #define farmalloc  malloc
   #undef  farrealloc
   #define farrealloc realloc
   #undef  farfree
   #define farfree    free
#else
   #define COMMAND_ID(wp,lp)        (wp)
   #define CONTROL_ID(wp,lp)        (wp)
   #define NOTIFY_MSG(wp,lp)        HIWORD(lp)
#endif

/******************************************************************************
    DLL and demo program specific includes
*******************************************************************************/
#include "rep.h"
#include "demo.h"

#include "util.h"      // need only by the demo program

/****************************************************************************
        Main Fuction
******************************************************************************/
int PASCAL WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,int nCmdShow)
{
    MSG msg;

    hPrevInst=hPrevInstance;
    if (!hPrevInstance)
        if (!InitApplication(hInstance))
            return (FALSE);

    if (!InitInstance(hInstance, nCmdShow))
        return (FALSE);

    while (GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    return (msg.wParam);
}

/****************************************************************************
   Initialize window data and register window class
****************************************************************************/

BOOL InitApplication(HANDLE hInstance)
{
    WNDCLASS  wc;

    wc.style = 0;
    wc.lpfnWndProc = (void far *)DemoWndProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = GetStockObject(WHITE_BRUSH);
    wc.lpszMenuName =  "DemoMenu";
    wc.lpszClassName = "ReDemoClass";

    return (RegisterClass(&wc));
}


/****************************************************************************
    Save instance handle and create main window
****************************************************************************/

BOOL InitInstance(HANDLE hInstance,int nCmdShow)
{
    int             i;

    hInst = hInstance;

    hDemoWin = CreateWindow(
        "ReDemoClass",
        "Report Ease Demo",
        WS_OVERLAPPEDWINDOW,
        0,
        0,
        400,
        300,
        NULL,
        NULL,
        hInstance,
        NULL
    );

    if (!hDemoWin)
        return (FALSE);

    ShowWindow(hDemoWin, nCmdShow);
    UpdateWindow(hDemoWin);

    //****** initialize the structure variables *********
    FormParm.open=FALSE;           // all windows closed in the beginning
    FormParm.x=50;                  // Initial X position of FORM edit window
    FormParm.y=50;                  // Initial Y position of FORM edit window
    FormParm.width=GetSystemMetrics(SM_CXSCREEN)-FormParm.x;        // Initial edit window width
    if (FormParm.width>1000) FormParm.width=1000;

    FormParm.height=(GetSystemMetrics(SM_CYSCREEN)*96)/100 - FormParm.y; // Initial edit window height
    FormParm.ShowMenu=TRUE;        // display menu
    FormParm.ShowVerBar=TRUE;      // display vertical scroll bar
    FormParm.ShowHorBar=TRUE;      // display horizontal scroll bar (N/A with word wrap)
    FormParm.ShowToolbar=TRUE;     // display the toolbar
    strcpy(FormParm.file,"");      // Text file name if the input type is 'F'
    FormParm.hInst=hInst;          // Current application instant handle
    FormParm.hPrevInst=hPrevInst;  // Previous application instant handle
    FormParm.hParentWnd=hDemoWin;  // let this be the parent of the editor window
    FormParm.style=WS_OVERLAPPEDWINDOW; //  Editor window style
    strcpy(FormParm.DataSetName,"CUSTOMER");
    //strcpy(FormParm.DataSetName,"");  // This statement would disable data-set check by the reporter

    //***** pass the pointer to the call back routines
    /** The field selection and verification routines must be included in the
        export section of your applications definition file **/

    FormParm.UserSelection=(USER_SELECTION)MakeProcInstance(UserFieldSelection,hInst); // field selection routine
    FormParm.VerifyField=(VERIFY_FIELD)MakeProcInstance(VerifyField,hInst);          // field verification routine
    if (!FormParm.UserSelection || !FormParm.VerifyField) return FALSE;

    //********** read field name for the files *******************
    strcpy(DataFile[0].name,"CUSTOMER");
    strcpy(DataFile[1].name,"SALES");

    for (i=0;i<MAX_FILES;i++) {
       if (!ReadFields(i)) return FALSE;
    }

    // load the customer logo bitmap
    hLogoBM=LoadBitmap(hInst,"LogoBM");
    Bitmap2DIB(hLogoBM,&hInfo,&pInfo,&hImage,&pImage);                      // extract bits from the bitmap

    // initialize the report viewer structure
    ViewParm.file[0]=0;             // let API prompt for file name selection
    ViewParm.device='A';            // ask user
    ViewParm.x=FormParm.x;          // specify the window coordinates for screen output
    ViewParm.y=FormParm.y;          // these fields needed only for screen output
    ViewParm.width=FormParm.width;
    ViewParm.height=FormParm.height;
    ViewParm.hInst=FormParm.hInst;
    ViewParm.hPrevInst=FormParm.hPrevInst;
    ViewParm.hParentWnd=FormParm.hParentWnd;
    ViewParm.style=FormParm.style;
    ViewParm.TextPitch=0;           // default text pitch
    strcpy(ViewParm.SwapDir,"");         // screen page swap location
    ViewParm.SuppressPrintMessages=FALSE;// display print time messages
    ViewParm.ShowToolbar=TRUE;      // show toolbar


    RepSetHdrPos(TRUE);
    RepSetDefPrinterFlag(TRUE);  // print the report to current printer - needed only for older report, or as a default value for new report being created

    return (TRUE);
}

/****************************************************************************
    Process Main Window messages
****************************************************************************/

long FAR PASCAL _export DemoWndProc(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam)
{
    LPVOID lpProcParam;


    switch (message) {
        case WM_COMMAND:
           switch (COMMAND_ID(wParam,lParam)) {
              case IDM_EDITOR:                 // open a new window
                CallEditor();             // call the form editor
                break;

              case IDM_EXECUTER:              // run a form template
                RunReport();              // run the report
                break;

              case IDM_VIEWER:           // view a report saved to a disk file
                ViewParm.file[0]=0;      // let RepView prompt for file name
                ViewParm.device='A';     // let RepView prompt for output device

                if (OneWindowMethod) {   // all reports viewed in one window
                    if (!hRepWnd) hRepWnd=CreateWindow(REP_CLASS,"",WS_OVERLAPPEDWINDOW|WS_VISIBLE|REP_NO_CLOSE,
                                            FormParm.x,FormParm.y,FormParm.width,FormParm.height,hDemoWin,
                                            NULL,hInst,NULL);

                    if (RepView2(hRepWnd,&ViewParm)!=0) return FALSE;
                }
                else {                   // create new window for every report
                    if (RepView(&ViewParm)!=0) return FALSE;// Intialize the report
                }

                break;

              case IDM_FEEDBACK:          // feedback to SubSystems, Inc.
                ShellExecute(hWnd,"open","http://www.subsystems.com/re_feedback1.htm",NULL,NULL,0);
                break;

              case IDM_OPTIONS:          // accept editor window options
                //RepSetDefPrinter("Brother MFC-9600","","");
                lpProcParam = MakeProcInstance(DemoOptions, hInst);
                DialogBox(hInst,"DemoOptions",hWnd,lpProcParam);
                FreeProcInstance(lpProcParam);
                break;

              case IDM_PRINTER:         // Set printer options for the next report
                SetPrinterProp=!SetPrinterProp; //toggle
                CheckMenuItem(GetMenu(hWnd),IDM_PRINTER,SetPrinterProp? MF_CHECKED : MF_UNCHECKED);
                break;

              case IDM_RTF_ATTACH:
                AttachRtfFiles=!AttachRtfFiles; //toggle
                CheckMenuItem(GetMenu(hWnd),IDM_RTF_ATTACH,AttachRtfFiles? MF_CHECKED : MF_UNCHECKED);
                break;

              case IDM_OTHER:             // quit demo
                DemoOther(FormParm.open?FormParm.hFrWnd:NULL);
                break;

              case IDM_EXIT:             // quit demo
                if (FormParm.open) {
                   if (!SendMessage(FormParm.hFrWnd,WM_QUERYENDSESSION,0,0L)) break;
                }
                PostQuitMessage(0);

              default:
                return (DefWindowProc(hWnd, message, wParam, lParam));
           }
           break;

        case WM_DESTROY:
            if (FormParm.open) {
               if (!SendMessage(FormParm.hFrWnd,WM_QUERYENDSESSION,0,0L)) break;
            }

            // release demo bitmap data
            if (hLogoBM) DeleteObject(hLogoBM);
            GlobalUnlock(hImage);
            GlobalUnlock(hInfo);
            GlobalFree(hImage);
            GlobalFree(hInfo);

            PostQuitMessage(0);
            break;

        /********************************************************************
                      Messages coming from the form editor Window
        *********************************************************************/
        case REP_CLOSE:                      // REP sends this message before closing
            FormParm.open=FALSE;             // mark Report Ease as closed
            break;

        case REP_PRINT_PROGRESS:
            //if (lParam==1) RepIgnoreCommand2(0,wParam);
            break;

        default:
            return (DefWindowProc(hWnd, message, wParam, lParam));
    }
    return (LRESULT)(NULL);
}

/****************************************************************************
    CallEditor:
    This function calls the form editor.
****************************************************************************/
CallEditor()
{
    int result;
    char string[80];

    if (GetFormSelection(FormParm.file,TRUE)>=0) {  // select a form file

       //*** let form designer use the default fonts ******
       FormParm.FontTypeFace[0]=0;    // default font type faces for report

       // A form can be loaded in two ways.

       // In the first method (below), you first create a blank
       // form window and then use the RepOpenForm function to open the form file into that window
       if (OneWindowMethod) {
           if (!hFormWnd) hFormWnd=CreateWindow(FR_CLASS,"",
                                   WS_OVERLAPPEDWINDOW|WS_VISIBLE|REP_MENU|REP_HSCROLL|REP_VSCROLL|REP_NO_CLOSE,
                                   FormParm.x,FormParm.y,FormParm.width,FormParm.height,hDemoWin,
                                   NULL,hInst,NULL);

           FormParm.hFrWnd=hFormWnd;
           RepOpenForm(hFormWnd,&FormParm);  // read the form file into the editor window
           FormParm.open=TRUE;
       }
       else {
           // in the second method, the form file is opened in its own window.  The RepEditForm
           // function first creates the form window and then reads the form file into it.

           result=RepEditForm(&FormParm);        // call the form editor
           if (result!=0) {               // print error if any
              wsprintf(string,"Error calling the form editor, code: %d",result);
              MessageBox(NULL,string,NULL,MB_OK);
           }
       }
       RepSetFlags(FormParm.hFrWnd,0,TRUE,RFLAG_ALT_NUM_DECIMAL);
    }

    return TRUE;
}

/****************************************************************************
    DemoOther:
    Run a test code.
****************************************************************************/
DemoOther(HWND hWnd)
{
   int result;
   //RepSetPrinter(hWnd, FALSE,"cust.fpc", FALSE);

   char lickey[25], licdate[13], compname[32];

       strcpy(compname, "Payroll Business Solutions Ltd");

       strcpy(lickey, "Y95US-38G7Z-RT392-01211");

       strcpy(licdate, "smo201211-1");

   result=RepSetLicenseInfo((unsigned char *)lickey, (unsigned char *)licdate, (unsigned char *)compname);
   OurPrintf("rsli, result: %d gli: %d",result,RepGetLicenseStatus());


   return TRUE;
}

/****************************************************************************
    DemoOptions:
    This dialog box allows the user to change the editor window options.
****************************************************************************/

BOOL FAR PASCAL _export DemoOptions(HWND hDlg,unsigned message,WPARAM wParam,LPARAM lParam)
{


    switch (message) {
        case WM_INITDIALOG:
            // Set initial values of the parameters
            SetDlgItemInt(hDlg,IDC_X,FormParm.x,0);
            SetDlgItemInt(hDlg,IDC_Y,FormParm.y,0);
            SetDlgItemInt(hDlg,IDC_WIDTH,FormParm.width,0);
            SetDlgItemInt(hDlg,IDC_HEIGHT,FormParm.height,0);
            SendMessage(GetDlgItem(hDlg,IDC_MENU),BM_SETCHECK, FormParm.ShowMenu, 0L);
            SendMessage(GetDlgItem(hDlg,IDC_VER_BAR),BM_SETCHECK, FormParm.ShowVerBar, 0L);
            SendMessage(GetDlgItem(hDlg,IDC_HOR_BAR),BM_SETCHECK, FormParm.ShowHorBar, 0L);
            SendMessage(GetDlgItem(hDlg,IDC_TOOLBAR),BM_SETCHECK, FormParm.ShowToolbar, 0L);
            SendMessage(GetDlgItem(hDlg,IDC_ONE_WINDOW_METHOD),BM_SETCHECK, OneWindowMethod, 0L);
            return (TRUE);

        case WM_COMMAND:
            switch (CONTROL_ID(wParam,lParam)) {
                case IDOK:
                   if (!GetNumVal(hDlg,IDC_X,&FormParm.x)) return (TRUE);
                   if (!GetNumVal(hDlg,IDC_Y,&FormParm.y)) return (TRUE);
                   if (!GetNumVal(hDlg,IDC_WIDTH,&FormParm.width)) return (TRUE);
                   if (!GetNumVal(hDlg,IDC_HEIGHT,&FormParm.height)) return (TRUE);
                   FormParm.ShowMenu=SendMessage(GetDlgItem(hDlg,IDC_MENU),BM_GETCHECK, 0, 0L);
                   FormParm.ShowVerBar=SendMessage(GetDlgItem(hDlg,IDC_VER_BAR),BM_GETCHECK, 0, 0L);
                   FormParm.ShowHorBar=SendMessage(GetDlgItem(hDlg,IDC_HOR_BAR),BM_GETCHECK, 0, 0L);
                   FormParm.ShowToolbar=ViewParm.ShowToolbar=SendMessage(GetDlgItem(hDlg,IDC_TOOLBAR),BM_GETCHECK, 0, 0L);
                   OneWindowMethod=SendMessage(GetDlgItem(hDlg,IDC_ONE_WINDOW_METHOD),BM_GETCHECK, 0, 0L);
                   EndDialog(hDlg, TRUE);
                   return (TRUE);

                case IDCANCEL:
                   EndDialog(hDlg, FALSE);
                   return (TRUE);
                default:
                   ;
            }
            break;
    }
    return (FALSE);
}

/******************************************************************************
    GetFormSelection:
    Shows available report forms and let user select one.
    In form edit mode, the user has two additional selections to
    create a new report form.
******************************************************************************/
GetFormSelection(char *CurFile,int FormEdit)
{
     int  select;
     LPVOID lpProcParam;

     TotalForms=0;

     strcpy(CurFile," ");

     #if defined (WIN32)
        GetFormFiles("*.FPC");          // get the report form files
     #else
        GetFormFiles("*.FP");          // get the report form files
     #endif

     if (FormEdit) {                // allow the user to create new forms
         NewReport=TotalForms;      // append the new report selection
         TotalForms++;
         strcpy(FormName[NewReport],"New Report Form");
         strcpy(FormFile[NewReport]," ");
     }
     else NewReport=TotalForms;

     //***** show the file seletion box and let user select a file *****
     lpProcParam = MakeProcInstance(DemoForms, hInst);
     select=DialogBox(hInst,"DemoForms",hDemoWin,lpProcParam);
     FreeProcInstance(lpProcParam);

     if (select>=0) {
        strcpy(CurFile,"");
        if (select!=NewReport) strcpy(CurFile,FormFile[select]);  // pass the file name
     }

     return select;
}

/****************************************************************************
    DemoForms:
    This dialog box allows the user to select a form file.
****************************************************************************/

BOOL FAR PASCAL _export DemoForms(HWND hDlg,unsigned message,WPARAM wParam,LPARAM lParam)
{
    int i,idx;

    switch (message) {
        case WM_INITDIALOG:
            // Set initial values of the parameters
            for (i=0;i<TotalForms;i++) {
               idx=(int)SendMessage(GetDlgItem(hDlg,IDC_FILE_BOX),LB_ADDSTRING, 0, (DWORD)(LPSTR)FormName[i]);
               SendMessage(GetDlgItem(hDlg,IDC_FILE_BOX),LB_SETITEMDATA, idx, i);
            }
            SendMessage(GetDlgItem(hDlg,IDC_FILE_BOX),LB_SETCURSEL, 0,0L);
            SetFocus(GetDlgItem(hDlg,IDC_FILE_BOX));
            return (FALSE);

        case WM_COMMAND:
            switch (CONTROL_ID(wParam,lParam)) {
                case IDC_FILE_BOX:
                    if (NOTIFY_MSG(wParam,lParam)!=LBN_DBLCLK) break;

                case IDOK:
                   i=SendMessage(GetDlgItem(hDlg,IDC_FILE_BOX),LB_GETCURSEL, 0,0L);     // index into the list box
                   i=SendMessage(GetDlgItem(hDlg,IDC_FILE_BOX),LB_GETITEMDATA, i,0L);   // index into our table
                   EndDialog(hDlg, i);
                   return (TRUE);

                case IDCANCEL:
                   EndDialog(hDlg, -1);
                   return (TRUE);
                default:
                   ;
            }
            break;
    }
    return (FALSE);
}

/****************************************************************************
    DemoFileField:
    This dialog box allows the user to select a data file or to select
    a field for a data file.  When lParam is -1, the dialog box shows
    the data files to select from.  When lParam is zero or greater than
    zero, the dialog box shows the fields for the file to select from.  The
    lParam, then specifies the index of the data file to choose the
    fields from.
****************************************************************************/

int FAR PASCAL _export DemoFileField(HWND hDlg,unsigned message,WPARAM wParam,LPARAM lParam)
{
    int i;

    switch (message) {
        case WM_INITDIALOG:
            // Set initial values of the parameters
            if (lParam<0) {                            // stuff the list box with data file names
               SetWindowText(hDlg,"Select Data File"); // set the header
               for (i=0;i<MAX_FILES;i++) {
                  SendMessage(GetDlgItem(hDlg,IDC_FILE_BOX),LB_ADDSTRING, 0, (DWORD)(LPSTR)DataFile[i].name);
               }
            }
            else {                                     // stuff the list box with data field names
               if (GetSortField) SetWindowText(hDlg,"Select Sort Field");// set the header
               else              SetWindowText(hDlg,"Select Data Field");// set the header
               for (i=0;i<DataFile[lParam].FieldCount;i++) {
                  SendMessage(GetDlgItem(hDlg,IDC_FILE_BOX),LB_ADDSTRING, 0, (DWORD)(LPSTR)DataField[lParam][i].ShortName);  // load the field names
               }
            }
            SendMessage(GetDlgItem(hDlg,IDC_FILE_BOX),LB_SETCURSEL, 0,0L);
            SetFocus(GetDlgItem(hDlg,IDC_FILE_BOX));
            return (FALSE);

        case WM_COMMAND:
            switch (CONTROL_ID(wParam,lParam)) {
                case IDC_FILE_BOX:
                    if (NOTIFY_MSG(wParam,lParam)!=LBN_DBLCLK) break;

                case IDOK:
                   i=SendMessage(GetDlgItem(hDlg,IDC_FILE_BOX),LB_GETCURSEL, 0,0L);
                   EndDialog(hDlg, i);
                   return (TRUE);

                case IDCANCEL:
                   EndDialog(hDlg, -1);
                   return (TRUE);
                default:
                   ;
            }
            break;
    }
    return (FALSE);
}

/******************************************************************************
    GetFormFiles:
    This routine scans the current directory for the files matching the
    specified template.
******************************************************************************/
GetFormFiles(char *template)
{
    unsigned done;
    FILE *iStream;
    struct StrFormHdr hdr;

    #if defined (WIN32)
       HANDLE hFindFile;
       WIN32_FIND_DATA blk;
       done=FALSE;
       if (INVALID_HANDLE_VALUE==(hFindFile=FindFirstFile(template,&blk))) done=TRUE;
    #else
       #if defined(__TURBOC__)
          struct ffblk blk;
          done=findfirst(template,&blk,FA_RDONLY);
       #else
          struct find_t blk;
          done=_dos_findfirst(template,_A_RDONLY,&blk);
       #endif
    #endif


    while(!done) {
       if (TotalForms==MAX_FORMS) {
          return FALSE;
       }

       #if defined (WIN32)
          strcpy(FormFile[TotalForms],blk.cFileName);
          done=!FindNextFile(hFindFile,&blk);
       #else
          #if defined(__TURBOC__)
             strcpy(FormFile[TotalForms],blk.ff_name);
             done=findnext(&blk);         // get the next file
          #else
             strcpy(FormFile[TotalForms],blk.name);
             done=_dos_findnext(&blk);
          #endif
       #endif

       //*************** get the form name ****************************
       if (NULL==(iStream=fopen(FormFile[TotalForms],"rb"))) continue;

       if (fread(&hdr,sizeof(struct StrFormHdr),1,iStream)!=1
          || hdr.FormSign!=FORM_SIGN ) {
           fclose(iStream);
           continue;
       }
       fclose(iStream);

       strcpy(FormName[TotalForms],hdr.name);

       TotalForms++;
    }

    return TRUE;
}

/******************************************************************************
    ReadFields:
    Read the field definitions for a file.  The argument specifies the index
    into the DataFile structure.
    The field definitions are stored in files with the extension .DF, whereas
    the data is stored in the files with the extension .DB.
******************************************************************************/
ReadFields(int idx)
{
    char name[64],line[300],CharReturn[300],*result;
    FILE *iStream;
    int  CurLen,CurField=0,LineIdx;
    int  ExtractField(char *,int *,int,char *);

    strcpy(name,DataFile[idx].name);
    strcat(name,".DF");              // file extension

    iStream=fopen(name,"r+t");
    if (iStream==NULL) {
        wsprintf(line,"Error while opening file %s ",(LPSTR)name);
        MessageBox(hDemoWin,line,NULL,MB_OK);
        return FALSE;
    }

    READ_LINE:
    result=fgets(line,300,iStream);
    if (ferror(iStream)) {
       wsprintf(line,"Error while reading file %s ",(LPSTR)name);
       MessageBox(hDemoWin,line,NULL,MB_OK);
       return FALSE;
    }
    if (feof(iStream))   goto END_FILE; // end of file
    if (result==NULL) {
       wsprintf(line,"Error while reading file %s ",(LPSTR)name);
       MessageBox(hDemoWin,line,NULL,MB_OK);
       return FALSE;
    }

    CurLen=strlen(line);             // exclude new line character
    if (CurLen>0 && line[CurLen-1]==0xA) CurLen--;
    line[CurLen]=0;

    LineIdx=0;                       // prepare to scan the line

    // extract the field name
    if (!ExtractField(line,&LineIdx,CurLen,CharReturn)) { // extract field name
       wsprintf(line,"Invalid format, file: %s, field# %d\n",(LPSTR)name,1);
       MessageBox(hDemoWin,line,NULL,MB_OK);
       return FALSE;
    }
    StringTrim(CharReturn);              // trim space
    strcpy(DataField[idx][CurField].ShortName,CharReturn);
    strcpy(DataField[idx][CurField].FullName,DataFile[idx].name);
    strcat(DataField[idx][CurField].FullName,"->");
    strcat(DataField[idx][CurField].FullName,CharReturn);

    // extract the field width
    if (!ExtractField(line,&LineIdx,CurLen,CharReturn)) { // extract field name
       wsprintf(line,"Invalid format, file: %s, field# %d\n",(LPSTR)name,1);
       MessageBox(hDemoWin,line,NULL,MB_OK);
       return FALSE;
    }
    DataField[idx][CurField].width=atoi(CharReturn);

    // extract the field type
    if (!ExtractField(line,&LineIdx,CurLen,CharReturn)) { // extract field name
       wsprintf(line,"Invalid format, file: %s, field# %d\n",(LPSTR)name,1);
       MessageBox(hDemoWin,line,NULL,MB_OK);
       return FALSE;
    }
    if      (CharReturn[0]=='T') DataField[idx][CurField].type=TYPE_TEXT;
    else if (CharReturn[0]=='N') DataField[idx][CurField].type=TYPE_NUM;
    else if (CharReturn[0]=='F') DataField[idx][CurField].type=TYPE_DBL;
    else if (CharReturn[0]=='D') DataField[idx][CurField].type=TYPE_DATE;
    else if (CharReturn[0]=='L') DataField[idx][CurField].type=TYPE_LOGICAL;
    else if (CharReturn[0]=='P') DataField[idx][CurField].type=TYPE_PICT;
    else {
       wsprintf(line,"Invalid field type, file: %s, field# %d\n",(LPSTR)name,1);
       MessageBox(hDemoWin,line,NULL,MB_OK);
       return FALSE;
    }

    // extract the default decimal places
    if (!ExtractField(line,&LineIdx,CurLen,CharReturn)) { // extract field name
       wsprintf(line,"Invalid format, file: %s, field# %d\n",(LPSTR)name,1);
       MessageBox(hDemoWin,line,NULL,MB_OK);
       return FALSE;
    }
    DataField[idx][CurField].DecPlaces=atoi(CharReturn);

    CurField++;

    goto READ_LINE;

    END_FILE:
    fclose(iStream);
    DataFile[idx].FieldCount=CurField;

    return TRUE;
}

/******************************************************************************
    ExtractField:
    This routine scans a text line to extract the next field.
    The fields are assumed to be of character type.
******************************************************************************/
ExtractField(char *line,int *idx,int LineLen,char *CharReturn)
{
    char quote='"',comma=',',string[32000],*ptr,SaveChar;             // string inclreased from 300 to 32000 to support rtf field
    int  i,j;

    i=(*idx);
    while (i<LineLen && line[i]==' ') i++;  // go past the spaces

    if (i>=LineLen) return FALSE;           // past the end of the line

    if (line[i]==quote) {
       i++;                                 // skip over the first quote
       if (NULL==(ptr=memchr(&line[i],quote,LineLen-i))) return FALSE;

       j=1;
       while (ptr[j]!=',' && ptr[j]!=0) j++;// locate the next comma

       SaveChar=ptr[0];                     // extract the string
       ptr[0]=0;
       strcpy(string,&line[i]);
       ptr[0]=SaveChar;
       (*idx)=i+strlen(string)+1+j;         // index to the next field
    }
    else {
       if (NULL==(ptr=memchr(&line[i],comma,LineLen-i))) ptr=&line[LineLen];  // field spans the end of the line

       SaveChar=ptr[0];                     // extract the string
       ptr[0]=0;
       strcpy(string,&line[i]);
       ptr[0]=SaveChar;
       (*idx)=i+strlen(string)+1;           // index to the next field
    }

    strcpy(CharReturn,string);
    return TRUE;
}


/******************************************************************************
    Call back routines.
    The following routines are called by Report Ease to select an application
    field or to verify an application field.
******************************************************************************/

/******************************************************************************
    UserFieldSelection:
    This routine is called by the FORM_FLD module to allow user to select
    a data field.  This routine can be programmed by your application in
    any way as long as it returns the data about the selected field using
    the argument pointer.  If the user chose not to select a field after all,
    the function should return with a FALSE value. Otherwise it should return
    with a TRUE value.

    In this routine, we will allow the user to first select a file, and
    then select a field from the chosen file.  The required data about the
    selected field is filled into the 'field' structure. Your program may
    also optionally fill other remaining 'field' structure variables.

    The first parameter contains the handle of the Form Editor window.
    The third parameter specifies the sort field number if this field will
    be used for section breaks. If your application will not sort on all
    fields,  you can restrict the fields that can be selected by the user.
    For non-sort fields, this paramter is set to 0.  In this demo program,
    we will allow only the fields from the primary file (CUSTOMER) to be
    used as sort fields.
******************************************************************************/
int FAR PASCAL _export UserFieldSelection(HWND hWnd,struct StrField huge *field,int SortFieldNo)
{
    int CurFile,CurField;
    FARPROC lpProc;

    SELECT_FILE:
    if (SortFieldNo==0) {
        lpProc = MakeProcInstance(DemoFileField, hInst);
        CurFile=DialogBoxParam(hInst,"DemoFileField",hWnd,(LPVOID)lpProc,(long)-1);
        FreeProcInstance(lpProc);           // set back the old window
        if (CurFile<0) return FALSE;
    }
    else CurFile=0;                         // Allow sort fields only from the customer file

    // SELECT FIELD
    if (SortFieldNo==0) GetSortField=FALSE; // global flag
    else                GetSortField=TRUE;

    lpProc = MakeProcInstance(DemoFileField, hInst);
    CurField=DialogBoxParam(hInst,"DemoFileField",hWnd,(LPVOID)lpProc,(long)CurFile);
    FreeProcInstance(lpProc);              // set back the old window

    if (CurField<0) {
       if (GetSortField) return FALSE;
       else              goto SELECT_FILE;
    }

    // fill the required 'field' variables

    lstrcpy(field->name,DataField[CurFile][CurField].FullName); // field name
    field->type=DataField[CurFile][CurField].type;             // alpha/num etc
    field->width=DataField[CurFile][CurField].width;           // display width
    field->DecPlaces=DataField[CurFile][CurField].DecPlaces;   // decimal places for display

    // specify new paragraph indicator field.  Used only for a word/wrapped
    // text type fields.
    field->ParaChar[0]='|';

    /****** fill up these OPTIONAL fields also. This information will be
            used by our report executer demo program to identfy the
            fields quickly.  *****/

    field->FileId=CurFile;                    // Information only
    field->FieldId=CurField;                  // information only

    return TRUE;
}

/******************************************************************************
    VerifyField:
    This routine is called by the FORM1 module to validate a field.  The field
    name is given by the 'name' variable in the StrField structure.  It
    contains the full name including the file prefix (if your application allows
    it). The input field is always in the upper case. The required data about the
    current field is filled into the 'field' structure. Your program may
    also optionally fill other remaining 'field' structure variables.

    The second argument indicates if the field can be
    used as a sort field.  This parameter indicates the sort field number.
    1 indicates the first sort field, 2 the second, ... A zero for this
    field indicates a non-sort field.  In this demo program, we will allow
    only the fields from the primary file (CUSTOMER) to be used as sort fields.

    The function returns TRUE if the field is valid, otherwise it returns
    a FALSE value.
******************************************************************************/
int FAR PASCAL _export VerifyField(HWND hWnd, struct StrField huge *field,int SortFieldNo)
{
    int  CurFile,CurField,MaxFiles;

    if (SortFieldNo==0) MaxFiles=MAX_FILES;
    else                MaxFiles=1;         // allow sort fields from the customer file only

    for (CurFile=0;CurFile<MaxFiles;CurFile++) {
       for (CurField=0;CurField<DataFile[CurFile].FieldCount;CurField++) {
         if (lstrcmp(DataField[CurFile][CurField].FullName,field->name)==0) {
            //* fill up the mandatory data *
            field->type=DataField[CurFile][CurField].type;             // alpha/num etc
            field->width=DataField[CurFile][CurField].width;           // display width
            field->DecPlaces=DataField[CurFile][CurField].DecPlaces;   // decimal places for display

            /****** fill up these OPTIONAL fields also. This information will be
                    used by our report executer demo program to identfy the
                    fields quickly.  *****/

            field->FileId=CurFile;                    // Information only
            field->FieldId=CurField;                  // information only
            return TRUE;
         }
       }
    }


    return FALSE;                                     // not a valid field
}

/******************************************************************************
    DrawPicture:
    This routine is called by the report executor to draw a specified
    picture type field.  The first argument specifies the device context
    of the reporting device.  If the report output is directed to a printer,
    this device context belongs to a printer, otherwise it specifies a
    device context for a metafile.  You application should draw the picture
    on this device context within the specified rectangle (last argument).

    The device context is in ANISOTROPIC mode.  The resolution in the X and
    Y direction is given by the UNITS_PER_INCH constant.

    The second argument is the picture id.  The third and fourth arguments
    are the file and field id for the picture field.

    The function returns TRUE if successful.

    In this demo program, this routine draws a logo for a given picture id.
    This program uses only one bitmap, and draws different part of the same
    bitmap for different picture id.

******************************************************************************/
int FAR PASCAL _export DrawPicture(HDC hDC, int PictId, int FileId, int FieldId, RECT far * rect, int ReportId, UINT flags)
{
    BITMAP bm;

    if (FieldId==11) {  // FieldId 11 in the customer file corresponds to 'logo' drawing
       #if defined(CHARTPRO_INTERFACE)   // show an example of ChartPro interface
          struct StrBarHdr BarHdr;
          int UnitsPerInch=250;  // ReportEase internal scale
          //HDC hMemDC;
          //HBITMAP hOldBM;
          //BITMAP bm;
          HGLOBAL  hInfo,hImage;              // handle to the device independent bitmap
          LPBYTE    pImage;                    // pointer to the image data
          LPBITMAPINFOHEADER pInfo;           // pointer to the image header data

          if (!RepIsPrinting(ReportId)) return TRUE;  // in preprocess


          //****** Initialize the StrBarHdr structure variables *********
          InitBarHdr(&BarHdr);       /* assign the default values to the structure
                                        variable.

                                        This step allows you to change only the
                                        variables that you would like to have
                                        non-default values.*/

          BarHdr.hDC     =GetDC(NULL);  /* Display context */
          BarHdr.hInst   =hInst;        // Handle of the current instanaces.


          BarHdr.UnitType=UNIT_LOGICAL;  // use logical units to convert pixels to mm

          BarHdr.ReturnBitmap=TRUE;      // return the output in a bitmap
          BarHdr.show=FALSE;             // do not show the bitmap on the screen

          //** Chart Size Parameters **
          BarHdr.width=rect->right-rect->left;   // Width in ReportEase units
          BarHdr.height=rect->bottom-rect->top;  // Height in report units

          BarHdr.width=MulDiv(BarHdr.width,254,UnitsPerInch*10); // convert to mm
          BarHdr.height=MulDiv(BarHdr.height,254,UnitsPerInch*10); // convert to mm

          ShowBar(&BarHdr);                      // create chart in a bitmap

          // now stretch blast to report DC
          GetObject(BarHdr.hBM,sizeof(BITMAP),(LPBYTE)&bm);     // get the dimension of bit map

          //hMemDC=CreateCompatibleDC(BarHdr.hDC);
          //hOldBM=SelectObject(hMemDC,BarHdr.hBM);
          //StretchBlt(hDC,rect->left,rect->top,rect->right-rect->left,rect->bottom-rect->top,
          //               hMemDC,0,0,bm.bmWidth,bm.bmHeight,SRCCOPY);
          // release resources
          //SelectObject(hMemDC,hOldBM);      // deselect chart bitmap

          // extract bitmap bits and copy to the device context
          Bitmap2DIB(BarHdr.hBM,&hInfo,&pInfo,&hImage,&pImage);                      // extract bits from the bitmap
          if (RepIsPrinting(ReportId))
               StretchDIBits(hDC,rect->left,rect->top,rect->right-rect->left,rect->bottom-rect->top,
                         0,0,bm.bmWidth,bm.bmHeight,pImage,(LPBITMAPINFO)pInfo,DIB_RGB_COLORS,SRCCOPY);

          GlobalUnlock(hInfo);
          GlobalFree(hInfo);
          GlobalUnlock(hImage);
          GlobalFree(hImage);

          DeleteObject(BarHdr.hBM);         // delete output bitmap
          //DeleteObject(hMemDC);             // delete temporary DC

          ReleaseDC(NULL,BarHdr.hDC);       // delete temporary windows dc

       #else                             // simply draw fragments of a bitmap
          int ResX,ResY;
          HDC hTempDC;
          int SourceWidth,SourceX,SourceHeight,DestWidth,DestHeight;
          RECT DrawRect;
          float FactorX,FactorY;

          // get the bitmap information
          GetObject(hLogoBM,sizeof(BITMAP),&bm);

          // Select part of the bitmap to display for this picture id
          SourceWidth=bm.bmWidth/10;          // divide picture into 10 equal parts
          if (PictId<1)  PictId=1;
          if (PictId>10) PictId=10;
          SourceX=SourceWidth*(PictId-1);

          SourceHeight=bm.bmHeight;

          // build the target rectangle
          DrawRect=(*rect);
          DestWidth=DrawRect.right-DrawRect.left;
          DestHeight=DrawRect.bottom-DrawRect.top;

          if (flags&FLAG_AUTOSIZE) {
             if (flags&FLAG_KEEP_ASPECT) {
                 FactorX=(float)SourceWidth/(float)DestWidth;  // how much the souce width is bigger
                 FactorY=(float)SourceHeight/(float)DestHeight;  // how much the souce height is bigger

                 if (FactorX>=FactorY) DestHeight=(int)(SourceHeight/FactorX);
                 else                  DestWidth=(int)(SourceWidth/FactorY);
             }
          }
          else {                             // draw picture in original size
             hTempDC=GetDC(NULL);
             ResX=GetDeviceCaps(hTempDC,LOGPIXELSX);
             ResY=GetDeviceCaps(hTempDC,LOGPIXELSY);
             ReleaseDC(NULL,hTempDC);

             DestWidth=MulDiv(SourceWidth,UNITS_PER_INCH,ResX);   // assume that the original picture was created for screen resolution
             DestHeight=MulDiv(SourceHeight,UNITS_PER_INCH,ResY);
          }

          // copy the bitmap - don't copy it when the reporter is in the preprocess mode
          if (RepIsPrinting(ReportId))
               StretchDIBits(hDC,DrawRect.left,DrawRect.top,DestWidth,DestHeight,
                         SourceX,0,SourceWidth,SourceHeight,pImage,(LPBITMAPINFO)pInfo,DIB_RGB_COLORS,SRCCOPY);

          rect->bottom=rect->top+DestHeight;   // pass the actual bottom of the picture

       #endif
    }
    else if (FieldId==12) { // FieldId 12 in the customer file corresponds to printing a letter using TER Edit Control
       if (PictId==1) RepDrawTerText(ReportId,"letter1.rtf",NULL,0,NULL,NULL);
       else           RepDrawTerText(ReportId,"letter2.rtf",NULL,0,NULL,NULL);
    }

    return TRUE;
}


/******************************************************************************
    Report Executer Interface Routines.
******************************************************************************/

/******************************************************************************
    RunReport: The following routines demonstrate the process of calling the
    report executor.
******************************************************************************/
RunReport()
{
   BOOL AppendReports=FALSE;
   BOOL result;

   struct StrRep  RepParm;            // parameter structure used to call the RepInit routine

   memset(&RepParm,0,sizeof(RepParm));

   //RepSetPdfInfo("test author","test producer","test subject","test keywords","","");  // 20171213

   if (GetFormSelection(RepParm.file,FALSE)>=0) { // select a form file

       //********* Initialize the argument parameters *************
       RepParm.device='A';            // ask user

       RepParm.x=FormParm.x;          // specify the window coordinates for screen output
       RepParm.y=FormParm.y;          // these fields needed only for screen output
       RepParm.width=FormParm.width;
       RepParm.height=FormParm.height;
       RepParm.hInst=FormParm.hInst;
       RepParm.hPrevInst=FormParm.hPrevInst;
       RepParm.hParentWnd=FormParm.hParentWnd;
       RepParm.style=FormParm.style;
       strcpy(RepParm.SwapDir,"");         // screen page swap location
       RepParm.SuppressPrintMessages=FALSE;// display print time messages
       RepParm.UseCurrentPrinter=FALSE;    // Let the 'report-parameter' control current printer selection
       RepParm.ShowToolbar=FormParm.ShowToolbar; // show toolbar
       RepParm.OutFile[0]=0;               // output file
       RepParm.TextPitch=0;                // default text pitch

       RepParm.DrawPicture=(DRAW_PICTURE)MakeProcInstance(DrawPicture,hInst);          // field verification routine
       if (!RepParm.DrawPicture) return FALSE;

       // optional method of setting printer properties before beginning the report
       if (SetPrinterProp) RepSetPrinter(FormParm.hParentWnd,FALSE,RepParm.file,FALSE);

       // A report can be initialized in two ways.
       // In the first method, you first create a report window and then initialize a report template
       // into it. The next report templates can be initialized into the same window. This method restricts
       // the report output to screen only.

       if (OneWindowMethod) {
           if (!hRepWnd) hRepWnd=CreateWindow(REP_CLASS,"",WS_OVERLAPPEDWINDOW|WS_VISIBLE|REP_NO_CLOSE,
                                   FormParm.x,FormParm.y,FormParm.width,FormParm.height,hDemoWin,
                                   NULL,hInst,NULL);

           if (RepInit2(hRepWnd,&RepParm)!=0) return FALSE;
       }
       else {
          // The second method calls the RepInit function to initialize the report.  This function
          // creates a report window automatically if the user wishes the report output to
          // go to a window.
          if (RepInit(&RepParm)!=0) return FALSE;// Intialize the report
       }
       RepId=RepParm.ReportId;

       if (AttachRtfFiles) AttachRtfDocs(RepParm.ReportId);  // example of attaching RTF files to the report

       PrepareFile(&RepParm);                // sort and join files if needed

       // do we need to do a preprocess pass
       if (RepParm.DoPreprocessPass) {
          RepBeginPreprocessPass(RepParm.ReportId);
          PrintRecords(RepParm.ReportId,RepParm.field,RepParm.TotalFields,FALSE);  // read and pass the records
          RepEndPreprocessPass(RepParm.ReportId);
       }

       PrintRecords(RepParm.ReportId,RepParm.field,RepParm.TotalFields,TRUE);  // read and print each record

       // Example of adding more reports to the initial report
       if (AppendReports) {
           strcpy(RepParm.file,"second.fpc");
           result=RepInitContinue(&RepParm);
           if (result==0) {
              PrepareFile(&RepParm);                // sort and join files if needed
              PrintRecords(RepParm.ReportId,RepParm.field,RepParm.TotalFields,TRUE);  // read and print each record
           }
       }

       RepExit(RepParm.ReportId);            // print footers and exit

       FreeProcInstance(RepParm.DrawPicture); // this callback no longer needed
   }

   return TRUE;
}

/******************************************************************************
    PrepareFile:  Sort and join the CUSTOMER and SALES files if needed.
******************************************************************************/
PrepareFile(struct StrRep far *RepParm)
{
    int i,SalesFileUsed,SortKey[10];
    char InputFile1[80],InputFile2[80],OutputFile[80];

    lstrcpy(InputFile1,DataFile[0].name);
    lstrcat(InputFile1,".DB ");              // apped the data file extension

    wsprintf(OutputFile,"temp%d.srt",RepParm->ReportId);

    for (i=0;i<RepParm->TotalSortFields;i++) {// extract each sort field
       SortKey[i]=RepParm->SortField[i].FieldId+1;
    }
    FileSort(InputFile1,OutputFile,RepParm->TotalSortFields,SortKey);
                         // the output is in the CUSTOMER.SRT file

    // determine if the sales file is used
    SalesFileUsed=FALSE;
    for (i=0;i<RepParm->TotalFields;i++) {
       if (RepParm->field[i].InUse && (RepParm->field[i].FileId)==1) { /* id =0 is customer file, and
                                                id=1 is the sales file,
                                                see the UserFieldSeletion routine */
           SalesFileUsed=TRUE;
           break;
       }
    }

    if (SalesFileUsed) {                     // join SALES file with CUSTOMER FILE
        lstrcpy(InputFile1,OutputFile);      // first file to join

        lstrcpy(InputFile2,DataFile[1].name);
        lstrcat(InputFile2,".DB ");          // second file

        FileJoin(InputFile1,1,InputFile2,1,OutputFile);

        /* customer and sales file have field number 1 in common. The output
        is stored in the "TEMPnnn.SRT file, where nnn is the ReportId
        */
    }

    return TRUE;
}

/******************************************************************************
    PrintRecords:
    This routine follows these step:
    1. Open the data set file.
    2. For each record of data set:
         a. Initialize the fields in the field structure.  IMPORTANT:  Initialize
            only the fields where source = SRC_APPL.  All other fields are
            for internal use of ReportEase.
         b. Parse each field from the data record and stuff into the
            field structure.  One field may be stuffed in more than one
            place in field structure.  Stuff only those structure fields where
            source = SRC_APPL
         c. Call the RepRec routine to print this record.

    This routine returns a TRUE value after all records are printed.  If
    the user hit escape during printing, then the return result will be a FALSE
    value.
******************************************************************************/
PrintRecords(int ReportId, struct StrField huge *field,int FieldCount,BOOL DelTempFile)
{
    FILE *iStream;
    int i,CurLen,FileNo,FieldNo,RecNo=1,LineIdx;
    char DataSetName[64],line[33000],string[32000],*result;       // string increased from 300 to 32000 to support rtf field

    //*** open the data set file ***
    wsprintf(DataSetName,"temp%d.srt",ReportId);  // The joined records are stored in the file TEMPnnn.SRT, where nnn is ReportId

    if (NULL==(iStream=fopen(DataSetName,"rt"))) {
       MessageBox(hDemoWin,"Can Not Open The Data Set File!",NULL,MB_OK);
       return FALSE;
    }


    READ_LINE:
    result=fgets(line,33000,iStream);
    if (ferror(iStream)) {
       MessageBox(hDemoWin,"Error in reading file",NULL,MB_OK);return FALSE;
    }
    if (feof(iStream))   goto END_FILE; // end of file
    if (result==NULL) {
       MessageBox(hDemoWin,"Error in reading file",NULL,MB_OK);return FALSE;
    }

    CurLen=strlen(line);               // exclude new line character
    if (CurLen>0 && line[CurLen-1]==0xA) CurLen--;
    line[CurLen]=0;

    //********** initialize the field structure *************
    for (i=0;i<FieldCount;i++) {
       if (field[i].source==SRC_APPL) {
          if (field[i].type==TYPE_TEXT)  field[i].CharData[0]=0;
          if (field[i].type==TYPE_NUM)   field[i].NumData=0;
          if (field[i].type==TYPE_PICT)  field[i].NumData=0;
          if (field[i].type==TYPE_DBL)   field[i].DblData=0;
          if (field[i].type==TYPE_DATE)  {    // the date field can be pass as numeric value, or double value
             field[i].NumData=0;
             field[i].DblData=0;
          }
       }
    }

    //************ parse the record to get each field ********
    FileNo=0;
    FieldNo=0;
    LineIdx=0;                         // prepare to scan the line

    while (TRUE) {
       // extract the fields
       if (!ExtractField(line,&LineIdx,CurLen,string)) break;

       if (FileNo>=MAX_FILES) {
          wsprintf(line,"Too many fields in the data record number: %d\n",RecNo);
          MessageBox(hDemoWin,line,NULL,MB_OK);
          goto END_FILE;
       }

       // stuff this field into the field structure
       for (i=0;i<FieldCount;i++) {
          if (field[i].source==SRC_APPL
                   && field[i].FileId==FileNo && field[i].FieldId==FieldNo) {
             if (field[i].type==TYPE_TEXT) {
                 BOOL done=FALSE;
                 if (strstr(string,"[RtfFile: ")==string) done=FillRtfData(ReportId,string,field,i);  // supply the rtf data from the specified rtf file

                 if (!done) {     // supply the text data from customer db file
                    if ((int)strlen(string)>field[i].width) RepEnlargeTextField(ReportId,i,strlen(string));
                    if ((int)strlen(string)>field[i].width) {
                       string[field[i].width]=0; // truncate oversize data
                    }
                    lstrcpy(field[i].CharData,string);
                 }
             }
             else if (field[i].type==TYPE_NUM)   field[i].NumData=atol(string);
             else if (field[i].type==TYPE_DBL)   field[i].DblData=(double)atof(string);
             else if (field[i].type==TYPE_DATE)  { // the date field can be pass as numeric value, or double value
                 //field[i].NumData=atol(string);    // the numeric date values should be passed as YYMMDD or YYYYMMDD
                 field[i].DblData=38008.7142592593; // example double value for 1/22/2004 5:08:32 PM - the double value should be compatible with the return value of the VisualBasic function CDbl(date)
             }
             else if (field[i].type==TYPE_PICT)  field[i].NumData=atol(string); // send a picture id
             else if (field[i].type==TYPE_LOGICAL) {
                if (string[0]=='Y' || string[0]=='y') field[i].NumData=1;
                else                                  field[i].NumData=0;
             }
          }
       }

       // advance to the next field number
       FieldNo++;
       if (FieldNo>=DataFile[FileNo].FieldCount) {
          FieldNo=0;
          FileNo++;
       }

    }

    if (RepRec(ReportId)!=0)  goto END_FILE;     // print this record

    RecNo++;

    goto READ_LINE;

    END_FILE:
    fclose(iStream);
    if (DelTempFile) unlink(DataSetName);

    return TRUE;
}

/******************************************************************************
    FillRtfData:
    Read the rtf data from a file called comment.rtf and pass to the specified
    text field.
******************************************************************************/
BOOL FillRtfData(int ReportId, LPSTR FileString, struct StrField *field, int FieldNo)
{
     BOOL done=FALSE;
     BYTE file[MAX_WIDTH+1];
     int len;
     DWORD size;
     HGLOBAL hMem;

     // the FileString contains the rtf file name in the [RtfFile: file.rtf] format
     lstrcpy(file,&(FileString[10]));  // file name starts after the '[RtfFile: ' FileString (10 characters)
     len=lstrlen(file);
     if (len>0) file[len-1]=0;     // disard the last

     hMem=RepFileToMem(file,&size);

     if (hMem!=NULL) {
       if (size>0) {
          LPSTR pMem=(LPSTR)GlobalLock(hMem);
          if (hMem!=NULL) {
             if ((int)size>field[FieldNo].width) RepEnlargeTextField(ReportId,FieldNo,size+1);
             if ((int)size<=field[FieldNo].width) {    // if RepEnlargeTextField succeeded, then field[FieldNo).width will now have updated width
                memmove(field[FieldNo].CharData,pMem,size);
                field[FieldNo].CharData[size]=0;

                GlobalUnlock(hMem);
                done=TRUE;
             }
          }
       }
       GlobalFree(hMem);
     }

     return done;
}

/******************************************************************************
    AttachRtfDocs:
    Example of attaching RTF files before and after the report.  The feature requires
    TE Developer's Kit installed with ReportEase Plus.
******************************************************************************/
AttachRtfDocs(int ReportId)
{
    BOOL continuous=FALSE;

    //RepSetRtfAttach2(ReportId,"attach.rtf",FALSE,"&DH_SoldToCompany|&DH_SoldToAddress2|&DH_SalesRep","Sub Systems, Inc.|3017 Covington Pl|Mike Smith",continuous);      // after the report

    //RepSetRtfAttach2(ReportId,"letter1a.rtf",FALSE,NULL,NULL,continuous);     // after the report
    //RepSetRtfAttach2(ReportId,"letter2a.rtf",FALSE,NULL,NULL,continuous);     // after the report

    RepSetRtfAttach2(ReportId,"attach1.rtf",TRUE,NULL,NULL,continuous);        // before the report

    //RepSetRtfAttach2(ReportId,"letter2.rtf",FALSE,NULL,NULL,continuous);     // after the report
    //RepSetRtfAttach2(ReportId,"letter1.rtf",FALSE,NULL,NULL,continuous);     // after the report
    //RepSetRtfAttach2(ReportId,"letter2.rtf",FALSE,NULL,NULL,continuous);     // after the report

    return TRUE;
}
/******************************************************************************
    Support Routines.
******************************************************************************/

/*****************************************************************************
    StringTrim:
    Remove spaces from the left and right of a NULL terminated string.
******************************************************************************/
void StringTrim(LPSTR string)
{
    RightTrim(string);LeftTrim(string);
}

/******************************************************************************
    RightTrim:
    Remove spaces on the right of a string.
******************************************************************************/
void RightTrim(LPSTR string)
{
    int i,TempLen;

    TempLen=lstrlen(string);
    for (i=TempLen-1;i>=0 && string[i] == ' ';i--);
    string[i+1] = '\0';
}

/******************************************************************************
    LeftTrim:
    Trim initial spaces from the string.
*******************************************************************************/
void LeftTrim(LPSTR string)
{
    int i,TempLen,BeginPoint;char TempStr[300];

    TempLen=lstrlen(string);
    for (i=0;i < TempLen && string[i] == ' ';i++);
    BeginPoint=i;
    for (i=BeginPoint;i<TempLen;i++) TempStr[i-BeginPoint]=string[i];
    TempStr[TempLen-BeginPoint] = '\0';
    lstrcpy(string,TempStr);
}

/******************************************************************************
     GetNumVal:
     Extract a numeric value from the dialog item.  Display error if a non-
     numeric value is encountered and return a false result.
*******************************************************************************/
BOOL GetNumVal(HWND hWin,int ControlId,int *pInt)
{
    BOOL ErrorFlag;

    *pInt=GetDlgItemInt(hWin,ControlId,&ErrorFlag,(BOOL) 0);
    if (!ErrorFlag) {
       MessageBox(hWin,"Must Enter a Valid Number",NULL,MB_OK);
       SetFocus(GetDlgItem(hWin,ControlId));
       return (FALSE);
    }
    return TRUE;
}

/******************************************************************************
     Bitmap2DIB:
     Extract device indepedent bits from the bitmap.
*******************************************************************************/
Bitmap2DIB(HBITMAP hBM, HGLOBAL far *phInfo, LPBITMAPINFOHEADER far *ppInfo,
    HGLOBAL far *phImage, LPBYTE far *ppImage)
{
    DWORD  InfoSize,ImageSize;
    BITMAP bm;
    HDC hDC;
    long ScanLineSize;
    HGLOBAL  hInfo,hImage;              // handle to the device independent bitmap
    LPBYTE    pImage;                    // pointer to the image data
    LPBITMAPINFOHEADER pInfo;           // pointer to the image header data


    GetObject(hBM,sizeof(BITMAP),(LPSTR)&bm);     // get the dimension of bit map

    //********** create the device independent bitmap ***********************
    InfoSize=sizeof(BITMAPINFOHEADER)+256*sizeof(RGBQUAD);  // create a 256 color bitmap structure

    if (NULL==(hInfo=GlobalAlloc(GMEM_MOVEABLE,InfoSize))
      ||NULL==(pInfo=(LPBITMAPINFOHEADER)GlobalLock(hInfo)) ) {
        MessageBox(hDemoWin,"Ran out of memory for Bitmap Info!",NULL,MB_OK);
        return -1;
    }

    //***** fill in the info structure ******
    pInfo->biSize=sizeof(BITMAPINFOHEADER);
    pInfo->biWidth=bm.bmWidth;
    pInfo->biHeight=bm.bmHeight;
    pInfo->biPlanes=1;                       // DIB have plane = 1
    pInfo->biBitCount=8;                     // 8 bits per pixel or color
    pInfo->biCompression=BI_RGB;             // no compression - needed for compatibility for all devices
    ScanLineSize=((pInfo->biWidth+sizeof(LONG)-1)/sizeof(LONG))*sizeof(LONG); // size of each scan line
    pInfo->biSizeImage=ScanLineSize*pInfo->biHeight; // buffer size requirement
    pInfo->biXPelsPerMeter=0;
    pInfo->biYPelsPerMeter=0;
    pInfo->biClrUsed=0;                      // depends on biBitCount
    pInfo->biClrImportant=0;                 // all colors important

    hDC=GetDC(hDemoWin);
    ImageSize=pInfo->biSizeImage;

    if (NULL==(hImage=GlobalAlloc(GMEM_MOVEABLE,ImageSize))
      ||NULL==(pImage=GlobalLock(hImage)) ) {
        MessageBox(hDemoWin,"Ran out of memory for Bitmap Image!",NULL,MB_OK);
        return -1;
    }

    if (!GetDIBits(hDC,hBM,0,bm.bmHeight,pImage,(LPBITMAPINFO)pInfo,DIB_RGB_COLORS)) {
        MessageBox(hDemoWin,"Error Creating Device Independent Bitmap Image!",NULL,MB_OK);
        return -1;
    }

    // release the common device context
    ReleaseDC(hDemoWin,hDC);

    // pass the bitmap infor
    if (phInfo) (*phInfo)=hInfo;
    if (ppInfo) (*ppInfo)=pInfo;
    if (phImage) (*phImage)=hImage;
    if (ppImage) (*ppImage)=pImage;

    return TRUE;
}

/******************************************************************************
    OurPrintf:
    This routine formats and display a given set of arguments.  The format
    is given by the first argument.  The argument 2 to n contain the
    arguments for the specified format.  The function uses MessageBox to
    display the formatted string.
******************************************************************************/
int OurPrintf(LPSTR fmt,...)
{
    LPSTR ArgStart;
    char string[300];
    HWND hWnd;

    ArgStart=(LPSTR) &fmt;  // pointer to first argument
    ArgStart=&ArgStart[4];  // go past the first argument
    wvsprintf(string,fmt,ArgStart);

    if (NULL!=(hWnd=FindWindow("DBWin",NULL))) { // debug window open
        if (NULL!=(hWnd=GetWindow(hWnd,GW_CHILD))) {    // get the edit window
           int len=lstrlen(string);
           string[len]=0xd;
           string[len+1]=0xa;
           string[len+2]=0;
           #if defined (WIN32)
              SendMessage(hWnd,EM_SETSEL,32767,32767);  // show the last line
              SendMessage(hWnd,EM_REPLACESEL,0,(DWORD)(LPSTR)string);
              SendMessage(hWnd,EM_SCROLLCARET,0,0);     // scroll caret into view
           #else
              SendMessage(hWnd,EM_SETSEL,0,MAKELPARAM(32767,32767));  // show the last line
              SendMessage(hWnd,EM_REPLACESEL,0,(DWORD)(LPSTR)string);
              SendMessage(hWnd,EM_SETSEL,0,MAKELPARAM(32767,32767));  // show the last line
           #endif
        }
    }
    else MessageBox(NULL,string,NULL,MB_OK);
    return TRUE;
}

